'''
HDRI Rotator Addon
Author: 1P2D - Felix Pierron (me@1p2d.com)
Website: https://1p2d.com

This addon allows you to rotate your HDRI or shadows in the viewport with CTRL+ALT+RMB.

License: GNU General Public License v3 (GPLv3)
'''

import bpy
import math
import mathutils

class HDRI_Rotator(bpy.types.Operator):
    """Rotate HDRI or Light Direction with your mouse"""
    bl_idname = "view3d.hdri_rotator"
    bl_label = "HDRI Rotator"
    bl_description = "Rotate HDRI or Light Direction with your mouse. Hold {SHIFT} for precision."
    bl_options = {'REGISTER', 'UNDO', 'BLOCKING'}

    MAX_RECURSION_DEPTH = 100

    def invoke(self, context, event):
        try:
            self.start_x = event.mouse_x
            self.start_y = event.mouse_y
            self.shift_held = event.shift

            # Store the window, space, and scene to use during modal execution
            self.window = context.window
            self.space = context.space_data
            self.scene = context.scene

            # Get a safe reference to the area
            self.area = self.get_area_safe()
            if not self.area:
                self.report({'WARNING'}, "Operator must be run in a 3D Viewport")
                return {'CANCELLED'}

            self.shading = self.space.shading
            self.shading_type = self.shading.type

            if self.shading_type == 'SOLID' and self.shading.show_shadows:
                # Solid shading mode with shadows enabled; rotate light direction
                self.mode = 'LIGHT_DIRECTION'
                self.start_light_direction = self.scene.display.light_direction.copy()
                # Set cursor icon for light direction rotation
                self.window.cursor_modal_set('SCROLL_XY')
            else:
                # Other shading modes; rotate HDRI
                # Get the use_scene_world setting from the shading settings
                self.use_scene_world = self.shading.use_scene_world_render
                if self.use_scene_world:
                    # For custom HDRI
                    self.rotation_input = self.find_rotation_input()
                    if self.rotation_input is not None:
                        default_value = self.rotation_input.default_value
                        if isinstance(default_value, (mathutils.Euler, mathutils.Vector)):
                            self.start_rotation = default_value.copy()
                        elif isinstance(default_value, (float, int)):
                            self.start_rotation = default_value
                        else:
                            self.report({'ERROR'}, "Unsupported rotation input type.")
                            self.cleanup()
                            return {'CANCELLED'}
                        self.mode = 'CUSTOM_HDRI_ROTATION'
                    else:
                        # Try to find a Mapping node connected to an Environment Texture node
                        self.mapping_node = self.find_mapping_node()
                        if self.mapping_node is None:
                            self.report({'WARNING'}, "No 'Rotation' input or Mapping node found in World nodes")
                            self.cleanup()
                            return {'CANCELLED'}
                        default_value = self.mapping_node.inputs['Rotation'].default_value
                        if isinstance(default_value, (mathutils.Vector, mathutils.Euler)):
                            self.start_rotation = default_value.copy()
                        else:
                            self.report({'ERROR'}, "Unsupported rotation input type in mapping node.")
                            self.cleanup()
                            return {'CANCELLED'}
                        self.mode = 'MAPPING_NODE_ROTATION'
                else:
                    # For built-in HDRI
                    self.start_rotation = self.get_builtin_hdr_rotation()
                    self.mode = 'HDRI_ROTATION'
                # Set cursor icon for HDRI rotation
                self.window.cursor_modal_set('SCROLL_X')

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}
        except Exception as e:
            self.report({'ERROR'}, f"Initialization failed: {e}")
            self.cleanup()
            return {'CANCELLED'}

    def modal(self, context, event):
        try:
            if event.type == 'MOUSEMOVE':
                shift_held = event.shift

                # Detect shift key state change
                if shift_held != self.shift_held:
                    self.shift_held = shift_held
                    # Update start_x and start_rotation to prevent jumps
                    self.start_x = event.mouse_x
                    self.start_y = event.mouse_y
                    if self.mode == 'LIGHT_DIRECTION':
                        self.start_light_direction = self.scene.display.light_direction.copy()
                    else:
                        self.start_rotation = self.get_current_rotation()

                delta_x = event.mouse_x - self.start_x
                delta_y = event.mouse_y - self.start_y

                precision_multiplier = 0.1 if self.shift_held else 1.0

                if self.mode == 'LIGHT_DIRECTION':
                    base_sensitivity = 0.005  # Adjust sensitivity as needed
                    sensitivity = base_sensitivity * precision_multiplier
                    rot_x = delta_y * sensitivity  # Vertical movement rotates around X axis
                    rot_z = -delta_x * sensitivity  # Horizontal movement rotates around Z axis

                    new_light_direction = self.calculate_new_direction(
                        self.start_light_direction,
                        rot_x,
                        rot_z
                    )
                    self.scene.display.light_direction = new_light_direction
                    self.safe_redraw()
                elif self.mode == 'CUSTOM_HDRI_ROTATION':
                    # Rotate custom HDRI based on mouse movement
                    base_sensitivity = 0.1  # Standard sensitivity
                    sensitivity = base_sensitivity * precision_multiplier

                    # Adjust sensitivity based on multiply value (for Polyhaven HDRIs)
                    multiply_factor = getattr(self, 'multiply_factor', 1.0)
                    sensitivity *= multiply_factor

                    delta_rotation = delta_x * sensitivity

                    if isinstance(self.start_rotation, mathutils.Euler):
                        new_rotation = self.start_rotation.copy()
                        new_rotation.z += math.radians(delta_rotation)
                    elif isinstance(self.start_rotation, mathutils.Vector):
                        new_rotation = self.start_rotation.copy()
                        new_rotation[2] += math.radians(delta_rotation)
                    elif isinstance(self.start_rotation, (float, int)):
                        new_rotation = self.start_rotation + math.radians(delta_rotation)
                    else:
                        self.report({'ERROR'}, "Unsupported rotation input type.")
                        self.cleanup()
                        return {'CANCELLED'}

                    self.set_custom_hdri_rotation(new_rotation)
                elif self.mode == 'MAPPING_NODE_ROTATION':
                    # Rotate custom HDRI via mapping node
                    base_sensitivity = 0.1  # Standard sensitivity
                    sensitivity = base_sensitivity * precision_multiplier
                    delta_rotation = delta_x * sensitivity

                    if isinstance(self.start_rotation, mathutils.Euler):
                        new_rotation = self.start_rotation.copy()
                        new_rotation.z += math.radians(delta_rotation)
                    elif isinstance(self.start_rotation, mathutils.Vector):
                        new_rotation = self.start_rotation.copy()
                        new_rotation[2] += math.radians(delta_rotation)
                    else:
                        self.report({'ERROR'}, "Unsupported rotation input type.")
                        self.cleanup()
                        return {'CANCELLED'}

                    self.set_mapping_node_rotation(new_rotation)
                elif self.mode == 'HDRI_ROTATION':
                    # Rotate built-in HDRI based on mouse movement
                    base_sensitivity = 0.1  # Standard sensitivity
                    sensitivity = base_sensitivity * precision_multiplier
                    delta_rotation = delta_x * sensitivity
                    new_rotation = self.start_rotation + delta_rotation
                    self.set_builtin_hdr_rotation(new_rotation)
                return {'RUNNING_MODAL'}
            elif event.type == 'RIGHTMOUSE' and event.value == 'RELEASE':
                # Finish the operation
                self.cleanup()
                return {'FINISHED'}
            elif event.type in {'ESC'}:
                # Cancel the operation
                # Restore original settings if necessary
                if self.mode == 'LIGHT_DIRECTION':
                    self.scene.display.light_direction = self.start_light_direction
                elif self.mode == 'CUSTOM_HDRI_ROTATION':
                    self.set_custom_hdri_rotation(self.start_rotation)
                elif self.mode == 'MAPPING_NODE_ROTATION':
                    self.set_mapping_node_rotation(self.start_rotation)
                elif self.mode == 'HDRI_ROTATION':
                    self.set_builtin_hdr_rotation(self.start_rotation)
                self.cleanup()
                return {'CANCELLED'}
            else:
                return {'RUNNING_MODAL'}
        except Exception as e:
            self.report({'ERROR'}, f"An error occurred: {e}")
            self.cleanup()
            return {'CANCELLED'}

    def cleanup(self):
        if hasattr(self, 'window') and self.window:
            self.window.cursor_modal_restore()
        # Reset any other stored states if necessary
        self.shift_held = False

    def safe_redraw(self):
        area = self.get_area_safe()
        if area:
            area.tag_redraw()

    def get_area_safe(self):
        # Safely get area reference
        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                return area
        return None

    def get_current_rotation(self):
        if self.mode == 'LIGHT_DIRECTION':
            return self.scene.display.light_direction.copy()
        elif self.mode == 'CUSTOM_HDRI_ROTATION':
            if isinstance(self.start_rotation, (mathutils.Euler, mathutils.Vector)):
                return self.rotation_input.default_value.copy()
            elif isinstance(self.start_rotation, (float, int)):
                return self.rotation_input.default_value
        elif self.mode == 'MAPPING_NODE_ROTATION':
            return self.mapping_node.inputs['Rotation'].default_value.copy()
        elif self.mode == 'HDRI_ROTATION':
            return self.get_builtin_hdr_rotation()

    def find_rotation_input(self, node_tree=None, visited=None, depth=0):
        if depth > self.MAX_RECURSION_DEPTH:
            self.report({'WARNING'}, "Maximum recursion depth reached in find_rotation_input()")
            return None
        if visited is None:
            visited = set()
        if node_tree is None:
            world = self.scene.world
            if world is None or not world.use_nodes:
                return None
            node_tree = world.node_tree

        if node_tree in visited:
            return None
        visited.add(node_tree)

        for node in node_tree.nodes:
            if node.type == 'GROUP':
                if 'Rotation' in node.inputs:
                    rotation_input = node.inputs['Rotation']
                    if rotation_input.is_linked or rotation_input.default_value is not None:
                        # Check inside the group node for Multiply nodes connected to the rotation input
                        self.multiply_factor = self.find_multiply_factor(node.node_tree, rotation_input, depth + 1)
                        return rotation_input
                # Recursively check inside the node group's node tree
                rotation_input = self.find_rotation_input(node.node_tree, visited, depth + 1)
                if rotation_input:
                    return rotation_input
            elif 'Rotation' in node.inputs:
                rotation_input = node.inputs['Rotation']
                if rotation_input.is_linked or rotation_input.default_value is not None:
                    return rotation_input
        return None

    def find_multiply_factor(self, node_tree, rotation_input, depth=0):
        if depth > self.MAX_RECURSION_DEPTH:
            self.report({'WARNING'}, "Maximum recursion depth reached in find_multiply_factor()")
            return 1.0  # Default multiply factor
        # Default multiply factor is 1.0
        multiply_factor = 1.0

        # Start from the input socket and traverse the node tree to find Multiply nodes
        nodes_to_visit = []
        visited_nodes = set()

        # Add all links connected to the rotation input
        for link in rotation_input.links:
            nodes_to_visit.append(link.from_node)

        while nodes_to_visit:
            node = nodes_to_visit.pop()
            if node in visited_nodes:
                continue
            visited_nodes.add(node)

            if node.type == 'MATH' and node.operation == 'MULTIPLY':
                # Get the multiply value
                if node.inputs[1].is_linked:
                    # If the multiply value is linked, we cannot determine it
                    self.report({'WARNING'}, "Cannot determine multiply value; assuming default sensitivity")
                    return multiply_factor
                multiply_value = node.inputs[1].default_value
                if multiply_value != 0:
                    multiply_factor /= multiply_value
                else:
                    self.report({'WARNING'}, "Multiply value is zero; assuming default sensitivity")
                    return multiply_factor
            elif node.type == 'COMBINEXYZ':
                # Continue traversing upstream
                for input_socket in node.inputs:
                    if input_socket.is_linked:
                        for link in input_socket.links:
                            nodes_to_visit.append(link.from_node)
            elif node.type == 'GROUP':
                # Check inside the group node
                group_multiply_factor = self.find_multiply_factor(node.node_tree, node.outputs[0], depth + 1)
                multiply_factor *= group_multiply_factor
            else:
                # Traverse upstream nodes
                for input_socket in node.inputs:
                    if input_socket.is_linked:
                        for link in input_socket.links:
                            nodes_to_visit.append(link.from_node)

        return multiply_factor

    def find_mapping_node(self):
        world = self.scene.world
        if world is None or not world.use_nodes:
            return None
        node_tree = world.node_tree

        visited_groups = set()
        nodes_to_check = list(node_tree.nodes)

        while nodes_to_check:
            node = nodes_to_check.pop()
            if node.type == 'GROUP':
                if node.node_tree and node.node_tree not in visited_groups:
                    visited_groups.add(node.node_tree)
                    nodes_to_check.extend(node.node_tree.nodes)
            elif node.type == 'MAPPING':
                # Check if this Mapping node is connected to an Environment Texture node
                if self.is_mapping_connected_to_env_texture(node):
                    return node
        return None

    def is_mapping_connected_to_env_texture(self, mapping_node):
        # Check downstream nodes to see if the mapping node ultimately connects to an Environment Texture
        nodes_to_check = [mapping_node]
        visited_nodes = set()

        while nodes_to_check:
            node = nodes_to_check.pop()
            visited_nodes.add(node)

            if node.type == 'TEX_ENVIRONMENT':
                return True

            for output in node.outputs:
                for link in output.links:
                    next_node = link.to_node
                    if next_node not in visited_nodes:
                        nodes_to_check.append(next_node)
                    if next_node.type == 'GROUP':
                        # Include nodes inside the group
                        if next_node.node_tree:
                            nodes_to_check.extend(next_node.node_tree.nodes)

        return False

    def set_custom_hdri_rotation(self, rotation_value):
        if isinstance(self.rotation_input.default_value, (mathutils.Euler, mathutils.Vector)):
            self.rotation_input.default_value = rotation_value
        elif isinstance(self.rotation_input.default_value, (float, int)):
            self.rotation_input.default_value = rotation_value
        else:
            self.report({'ERROR'}, "Unsupported rotation input type.")
            return
        # Force update of the scene to ensure the change takes effect
        bpy.context.view_layer.update()
        # Redraw the viewport
        self.safe_redraw()

    def set_mapping_node_rotation(self, rotation_value):
        if isinstance(self.mapping_node.inputs['Rotation'].default_value, mathutils.Vector):
            rotation = self.mapping_node.inputs['Rotation'].default_value.copy()
            if isinstance(rotation_value, mathutils.Vector):
                rotation[2] = rotation_value[2]
            else:
                rotation[2] = rotation_value
            self.mapping_node.inputs['Rotation'].default_value = rotation
        elif isinstance(self.mapping_node.inputs['Rotation'].default_value, mathutils.Euler):
            rotation = self.mapping_node.inputs['Rotation'].default_value.copy()
            if isinstance(rotation_value, mathutils.Euler):
                rotation.z = rotation_value.z
            else:
                rotation.z = rotation_value
            self.mapping_node.inputs['Rotation'].default_value = rotation
        elif isinstance(self.mapping_node.inputs['Rotation'].default_value, (float, int)):
            self.mapping_node.inputs['Rotation'].default_value = rotation_value
        else:
            self.report({'ERROR'}, "Unsupported rotation input type in mapping node.")
            return
        # Force update
        bpy.context.view_layer.update()
        self.safe_redraw()

    def get_builtin_hdr_rotation(self):
        # Get the built-in HDRI rotation in degrees
        rotation_rad = self.shading.studiolight_rotate_z
        rotation_deg = math.degrees(rotation_rad)
        return rotation_deg

    def set_builtin_hdr_rotation(self, rotation_value):
        # Set the built-in HDRI rotation
        rotation_rad = math.radians(rotation_value)
        # Wrap rotation to -pi to pi
        rotation_rad_wrapped = (rotation_rad + math.pi) % (2 * math.pi) - math.pi
        self.shading.studiolight_rotate_z = rotation_rad_wrapped
        self.safe_redraw()

    def calculate_new_direction(self, start_direction, rot_x, rot_z):
        # Rotate the start_direction vector around X and Z axes
        rot_x_mat = mathutils.Matrix.Rotation(rot_x, 4, 'X')
        rot_z_mat = mathutils.Matrix.Rotation(rot_z, 4, 'Z')

        # Apply rotations
        new_direction = rot_z_mat @ rot_x_mat @ start_direction

        # Normalize the vector
        if new_direction.length_squared != 0:
            new_direction.normalize()
        else:
            new_direction = mathutils.Vector((0.0, 0.0, -1.0))  # Default direction

        # Ensure the Z component is negative (light from above)
        if new_direction.z > 0:
            new_direction.z = -new_direction.z

        return new_direction

    def execute(self, context):
        self.cleanup()
        return {'FINISHED'}

    def cancel(self, context):
        # Restore original settings if necessary
        if self.mode == 'LIGHT_DIRECTION':
            self.scene.display.light_direction = self.start_light_direction
        elif self.mode == 'CUSTOM_HDRI_ROTATION':
            self.set_custom_hdri_rotation(self.start_rotation)
        elif self.mode == 'MAPPING_NODE_ROTATION':
            self.set_mapping_node_rotation(self.start_rotation)
        elif self.mode == 'HDRI_ROTATION':
            self.set_builtin_hdr_rotation(self.start_rotation)
        self.cleanup()
        return {'CANCELLED'}

def register():
    bpy.utils.register_class(HDRI_Rotator)

def unregister():
    bpy.utils.unregister_class(HDRI_Rotator)

if __name__ == "__main__":
    register()
